#!/usr/bin/env python
#
#####
# $Id: eventloop.py,v 1.4 1997/08/01 20:33:12 timo Exp $
#
#####

import select, time

class Event:
    """Event class
    A base class for different Event types."""
    def __init__(self, func=None, arg=None):
	self.func = func
	self.arg  = arg

    def __call__(self):
	apply(self.func, self.arg)
# end class Event


class FileEvent(Event):
    """A File Event
    We keep a reference to the file object.  File objects have a tendency
    to close when deleted, so this reference prevents the user from 
    accidentally closing the file without telling the EventLoop.
    """
    
    def __init__(self, file, func, arg):
	Event.__init__(self, func, arg)
	self.file = file
	self.fileno = file.fileno
	return
    # end __init__
# end class FileEvent


class TimerEvent(Event):
    """A Timer Event
    The only interesting thing here is the need to generate a unique-id
    for each timer.
    """
    
    next_id = 1
    last_id = 65535
    in_use  = []

    def __init__(self, expire, func, arg):
	Event.__init__(self, func, arg)
	self.expire = expire
	self.id = N = TimerEvent.next_id
	
	# Update in_use array
	TimerEvent.in_use.append(N)

	# Make sure the next_id is unique
	if N == TimerEvent.last_id:
	    N = 1
	else:
	    N = N + 1
	while N in TimerEvent.in_use:
	    N = N + 1
	TimerEvent.next_id = N + 1
	return
    # end __init__

    def __del__(self):
	TimerEvent.in_use.remove(self.id)
    # end __del__
# end class TimerEvent



LoopExit = "LoopExit"

class EventLoop:
    """EventLoop creates a select-based mainloop.  Events may be
    added as either timers, which 'fire' after a specified period,
    or as fileobjects, which "trigger" when available for reading.
    Once a timer fires, it is removed from the Eventloop.  If you
    want your timer to fire every N seconds, then you must call
    AddTimer from your callback.  (See example below.)
    A fileobject can be any object with a 'fileno' method.  A
    file event must be explicitly removed via DeleteFile.  In this
    respect, FileEvents differ from TimerEvents.
    """

    ###
    ###  Internally:
    ###
    ###      We have one structure to keep file objects:
    ###            __filelist  ==   { fileno : file_object }
    ###
    ###      We have one structure to keep timers:
    ###            __timerlist ==  { expiretime : [timer ...] }
    ###

    def __init__(self):
	self.__filelist  = {}
	self.__timerlist = {}
	return
    # end __init__

    def Cancel(self, val=None):
	raise LoopExit, val

    def MainLoop(self):
	sf = self.__filelist
	st = self.__timerlist

	try:
	    # while-loop implements the mainloop
	    while 1:

		delay = None  # Indicates no timers
		if st:
		    when = min( st.keys() )   # Next timer
		    now = time.time()         # Current time
		    if now >= when:
			self.FireTimers(when) # Fire any expired timers
			continue
		    # else
		    delay = when - now       # Delay until next timer

		# Although undocumented,
		#   delay==None results in a blocking select() call.
		r,w,e = select.select( sf.values(), [], [], delay)

		# For-Loop handles any file activity
		for File in r:
		    File()

	    # end while loop
        except LoopExit, val:
	    return val
	# should never reach here
	return
    # end MainLoop


    def AddFile(self, obj, callback, *arg):
	sf = self.__filelist
	f_no = obj.fileno()
	#   if sf.has_key(f_no):
	# 	raise ValueError, "FileNo already used!"
	sf[f_no] = FileEvent(obj, callback, arg)
	return
    # end AddFile

    def DeleteFile(self, obj):
	sf = self.__filelist
	f_no = obj.fileno()
	del sf[f_no]
	return
    # end DeleteFile

    def AddTimer(self, seconds, callback, *arg):
	st = self.__timerlist
	when = time.time() + seconds
	tobj = TimerEvent(when, callback, arg)
	if st.has_key(when):
	    st[when].append(tobj)
	else:
	    st[when] = [tobj]
	return tobj.id
    # end AddTimer

    def DeleteTimer(self, identifier):
	st = self.__timerlist
	for T in st.values():
	    for Timer in T:
		if Timer.id == identifier:
		    when = Timer.expire
		    T.remove(Timer)
		    if st[when] == []: del st[when]
		    return
	# We didn't find it
	raise ValueError, "No timer with identifier" + `identifier`
    # end DeleteTimer

    def FireTimers(self, which):
	st = self.__timerlist
	TL = st[which]
	del st[which]

	for Timer in TL:
	    Timer()
	return
    # end FireTimer


# end class EventLoop


def _timertest(endtime):
    global Ev
    t_minus = endtime - time.time()

    if t_minus <= 0:
	print "Boom!!!"
	Ev.Cancel()

    print
    print "You have", `int(t_minus) +1`, "seconds to comply."
    print
    Ev.AddTimer(10, _timertest, end)
# end _timertest

def _filetest(str, endtime):
    global Ev
    ans = sys.stdin.readline()
    
    if ans[:-1] == str:
	print "Hooray!"
	Ev.Cancel()

    print "Wrong!  Try again."
# end _filetest

if __name__ == "__main__":
    import sys
    global Ev
    phrase = "Python is my favorite language"

    Ev = EventLoop()

    print "Please type:"
    print "      ", phrase

    end = time.time() + 30
    Ev.AddFile(sys.stdin, _filetest, phrase, end )
    Ev.AddTimer(0, _timertest, end)
    Ev.MainLoop()
